@@ -1,4 +1,8 @@
-//  Copyright (C) 2001  MandrakeSoft S.A.
+/////////////////////////////////////////////////////////////////////////
+// $Id: win32.cc,v 1.30 2002/03/28 01:12:26 bdenney Exp $
+/////////////////////////////////////////////////////////////////////////
+//
+//  Copyright (C) 2002  MandrakeSoft S.A.
 //
 //    MandrakeSoft S.A.
 //    43, rue d'Aboukir
@@ -80,6 +84,9 @@ static RECT updated_area;
 static BOOL updated_area_valid = FALSE;
 static Bit32u *VideoBits;
 
+// Textmode cursor
+HBITMAP cursorBmp;
+
 // Headerbar stuff
 unsigned bx_bitmap_entries = 0;
 struct {
@@ -112,6 +119,9 @@ static unsigned prev_block_cursor_x = 0;
 static unsigned prev_block_cursor_y = 0;
 static HBITMAP vgafont[256];
 static unsigned x_edge=0, y_edge=0, y_caption=0;
+static int yChar = 16;
+static HFONT hFont[3];
+static int FontId = 2;
 
 static char szAppName[] = "Bochs for Windows";
 static char szWindowName[] = "Bochs for Windows - Display";
@@ -136,8 +146,11 @@ void terminateEmul(int);
 void create_vga_font(void);
 static unsigned char reverse_bitorder(unsigned char);
 void DrawBitmap (HDC, HBITMAP, int, int, DWORD, unsigned char cColor);
+void DrawChar (HDC, unsigned char, int, int, unsigned char cColor, int, int);
 void updateUpdated(int,int,int,int);
 static void headerbar_click(int x);
+void InitFont(void);
+void DestroyFont(void);
 
 
 /* Macro to convert WM_ button state to BX button state */
@@ -148,7 +161,6 @@ static void headerbar_click(int x);
   void bx_signal_handler(int);
 #endif
 
-
 static void processMouseXY( int x, int y, int windows_state, int implied_state_change)
 {
   int bx_state;
@@ -159,7 +171,7 @@ static void processMouseXY( int x, int y, int windows_state, int implied_state_c
   if ( old_bx_state!=mouse_button_state)
   {
     /* Make up for missing message */
-    BX_INFO(( "&&&missing mouse state change\r\n"));
+    BX_INFO(( "&&&missing mouse state change"));
     EnterCriticalSection( &stInfo.keyCS);
     enq_mouse_event();
     mouse_button_state=old_bx_state;
@@ -221,19 +233,20 @@ void terminateEmul(int reason) {
 
   for (unsigned c=0; c<256; c++)
     if (vgafont[c]) DeleteObject(vgafont[c]);
+  DeleteObject(cursorBmp);
 
   switch (reason) {
   case EXIT_GUI_SHUTDOWN:
-    BX_PANIC(("Window closed, exiting!\n"));
+    BX_PANIC(("Window closed, exiting!"));
     break;
   case EXIT_GMH_FAILURE:
-    BX_PANIC(("GetModuleHandle failure!\n"));
+    BX_PANIC(("GetModuleHandle failure!"));
     break;
   case EXIT_FONT_BITMAP_ERROR:
-    BX_PANIC(("Font bitmap creation failure!\n"));
+    BX_PANIC(("Font bitmap creation failure!"));
     break;
   case EXIT_HEADER_BITMAP_ERROR:
-    BX_PANIC(("Header bitmap creation failure!\n"));
+    BX_PANIC(("Header bitmap creation failure!"));
     break;
   case EXIT_NORMAL:
     break;
@@ -262,7 +275,7 @@ void terminateEmul(int reason) {
 void bx_gui_c::specific_init(bx_gui_c *th, int argc, char **argv, unsigned
 			     tilewidth, unsigned tileheight,
 			     unsigned headerbar_y) {
-  th->setprefix("[WGUI]");
+  th->put("WGUI");
   static RGBQUAD black_quad={ 0, 0, 0, 0};
   stInfo.kill = 0;
   stInfo.UIinited = FALSE;
@@ -277,13 +290,15 @@ void bx_gui_c::specific_init(bx_gui_c *th, int argc, char **argv, unsigned
 
   bx_headerbar_y = headerbar_y;
   dimension_x = 640;
-  dimension_y = 480 + bx_headerbar_y;
+  dimension_y = 800 + bx_headerbar_y;
   stretched_x = dimension_x;
   stretched_y = dimension_y;
   stretch_factor = 1;
 
   for(unsigned c=0; c<256; c++) vgafont[c] = NULL;
   create_vga_font();
+  cursorBmp = CreateBitmap(8,16,1,1,NULL);
+  if (!cursorBmp) terminateEmul(EXIT_FONT_BITMAP_ERROR);
 
   bitmap_info=(BITMAPINFO*)new char[sizeof(BITMAPINFOHEADER)+
     256*sizeof(RGBQUAD)];
@@ -324,8 +339,13 @@ void bx_gui_c::specific_init(bx_gui_c *th, int argc, char **argv, unsigned
   //  the emulated CPU runs, and it hogs the real CPU
   SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL);
 
-  if (bx_options.private_colormap)
-    BX_INFO(( "private_colormap option ignored.\n"));
+  if (bx_options.Oprivate_colormap->get ())
+    BX_INFO(( "private_colormap option ignored."));
+
+  // load keymap tables
+  if(bx_options.keyboard.OuseMapping->get()) {
+    bx_keymap.loadKeymap(NULL);  // I have no function to convert X windows symbols
+    }
 }
 
 
@@ -358,8 +378,8 @@ VOID UIThread(PVOID pvoid) {
 			      WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
 			      CW_USEDEFAULT,
 			      CW_USEDEFAULT,
-			      stretched_x + x_edge * 2,
-			      stretched_y + y_edge * 2 + y_caption,
+			      dimension_x + x_edge * 2,
+			      dimension_y + y_edge * 2 + y_caption,
 			      NULL,
 			      NULL,
 			      stInfo.hInstance,
@@ -367,6 +387,8 @@ VOID UIThread(PVOID pvoid) {
 
   if (stInfo.hwnd) {
     ShowWindow (stInfo.hwnd, SW_SHOW);
+    SetWindowPos (stInfo.hwnd, NULL, 0, 0, stretched_x + x_edge * 2,
+                  480 + y_edge * 2 + y_caption, SWP_NOMOVE|SWP_NOZORDER);
     UpdateWindow (stInfo.hwnd);
 
     ShowCursor(!mouseCaptureMode);
@@ -376,7 +398,7 @@ VOID UIThread(PVOID pvoid) {
     cursorWarped();
 
     hdc = GetDC(stInfo.hwnd);
-    MemoryBitmap = CreateCompatibleBitmap(hdc, dimension_x, dimension_y);
+    MemoryBitmap = CreateCompatibleBitmap(hdc, BX_MAX_XRES, BX_MAX_YRES);
     MemoryDC = CreateCompatibleDC(hdc);
     ReleaseDC(stInfo.hwnd, hdc);
 
@@ -404,27 +426,25 @@ LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam) {
 
   switch (iMsg) {
   case WM_CREATE:
+    InitFont();
     SetTimer (hwnd, 1, 330, NULL);
+    bx_options.Omouse_enabled->set (mouseCaptureMode);
     if (mouseCaptureMode)
-      SetWindowText(hwnd, "Bochs for Windows      [Press F12 to release mouse capture]");
+      SetWindowText(hwnd, "Bochs for Windows      [F12 to release mouse]");
     else
-      SetWindowText(hwnd, "Bochs for Windows      [F12 enables the mouse in Bochs]");
+      SetWindowText(hwnd, "Bochs for Windows      [F12 enables mouse]");
     return 0;
 
   case WM_TIMER:
     // If mouse escaped, bring it back
     if ( mouseCaptureMode)
     {
-      GetCursorPos( &ptCursorPos);
       GetWindowRect(hwnd, &wndRect);
-      if ( ptCursorPos.x<wndRect.left || ptCursorPos.x>wndRect.right
-        || ptCursorPos.y<( wndRect.top+BX_HEADER_BAR_Y*stretch_factor) || ptCursorPos.y>wndRect.bottom)
-      {
-        SetCursorPos(wndRect.left + stretched_x/2 + x_edge,
-                   wndRect.top + stretched_y/2 + y_edge + y_caption);
-        cursorWarped();
-      }
+      SetCursorPos(wndRect.left + stretched_x/2 + x_edge,
+                 wndRect.top + stretched_y/2 + y_edge + y_caption);
+      cursorWarped();
     }
+    bx_options.Omouse_enabled->set (mouseCaptureMode);
     
     return 0;
 
@@ -476,6 +496,7 @@ LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam) {
   case WM_DESTROY:
     KillTimer (hwnd, 1);
     stInfo.UIinited = FALSE;
+    DestroyFont();
 
     PostQuitMessage (0);
     return 0;
@@ -484,6 +505,7 @@ LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam) {
   case WM_SYSKEYDOWN:
     if (wParam == VK_F12) {
       mouseCaptureMode = !mouseCaptureMode;
+      bx_options.Omouse_enabled->set (mouseCaptureMode);
       ShowCursor(!mouseCaptureMode);
       ShowCursor(!mouseCaptureMode);   // somehow one didn't do the trick (win98)
       GetWindowRect(hwnd, &wndRect);
@@ -496,7 +518,7 @@ LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam) {
 	SetWindowText(hwnd, "Bochs for Windows      [F12 enables the mouse in Bochs]");
     } else {
       EnterCriticalSection(&stInfo.keyCS);
-      enq_key_event(HIWORD (lParam) & 0xFF, BX_KEY_PRESSED);
+      enq_key_event(HIWORD (lParam) & 0x01FF, BX_KEY_PRESSED);
       LeaveCriticalSection(&stInfo.keyCS);
     }
     return 0;
@@ -504,7 +526,7 @@ LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam) {
   case WM_KEYUP:
   case WM_SYSKEYUP:
     EnterCriticalSection(&stInfo.keyCS);
-    enq_key_event(HIWORD (lParam) & 0xFF, BX_KEY_RELEASED);
+    enq_key_event(HIWORD (lParam) & 0x01FF, BX_KEY_RELEASED);
     LeaveCriticalSection(&stInfo.keyCS);
     return 0;
 
@@ -521,7 +543,7 @@ LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam) {
 
 void enq_key_event(Bit32u key, Bit32u press_release) {	
   if (((tail+1) % SCANCODE_BUFSIZE) == head) {
-    BX_ERROR(( "enq_scancode: buffer full\n"));
+    BX_ERROR(( "enq_scancode: buffer full"));
     return;
   }
   keyevents[tail].key_event = key | press_release;
@@ -534,7 +556,7 @@ void enq_mouse_event(void)
   if ( ms_xdelta || ms_ydelta)
   {
     if (((tail+1) % SCANCODE_BUFSIZE) == head) {
-      BX_ERROR(( "enq_scancode: buffer full\n" ));
+      BX_ERROR(( "enq_scancode: buffer full" ));
       return;
     }
     QueueEvent& current=keyevents[tail];
@@ -552,7 +574,7 @@ QueueEvent* deq_key_event(void) {
   QueueEvent* key;
 
   if ( head == tail ) {
-    BX_ERROR(("deq_scancode: buffer empty\n"));
+    BX_ERROR(("deq_scancode: buffer empty"));
     return((QueueEvent*)0);
   }
   key = &keyevents[head];
@@ -603,6 +625,11 @@ void bx_gui_c::handle_events(void) {
       headerbar_click(LOWORD(key));
     }
     else {
+      if (((key & 0x0100) && ((key & 0x01ff) != 0x0145)) | ((key & 0x01ff) == 0x45)) {
+        // Its an extended key
+        scancode = 0xE0;
+        bx_devices.keyboard->put_scancode(&scancode, 1);
+      }
       // Its a key
       scancode = LOBYTE(LOWORD(key));
       // printf("# key = %d, scancode = %d\n",key,scancode);
@@ -676,26 +703,37 @@ void bx_gui_c::clear_screen(void) {
 
 void bx_gui_c::text_update(Bit8u *old_text, Bit8u *new_text,
 			   unsigned long cursor_x, unsigned long cursor_y,
-                           unsigned nrows) {
+                           Bit16u cursor_state, unsigned nrows) {
   HDC hdc;
   unsigned char cChar;
   unsigned i, x, y;
+  Bit8u cs_start, cs_end;
   unsigned nchars;
+  unsigned char data[32];
+
+  cs_start = (cursor_state >> 8) & 0x3f;
+  cs_end = cursor_state & 0x1f;
 
   if (!stInfo.UIinited) return;
-	
+
   EnterCriticalSection(&stInfo.drawCS);
 
   hdc = GetDC(stInfo.hwnd);
 
   // Number of characters on screen, variable number of rows
   nchars = 80*nrows;
-	
+
   if ( (prev_block_cursor_y*80 + prev_block_cursor_x) < nchars) {
     cChar = new_text[(prev_block_cursor_y*80 + prev_block_cursor_x)*2];
-    DrawBitmap(hdc, vgafont[cChar], prev_block_cursor_x*8,
-	              prev_block_cursor_y*16 + bx_headerbar_y, SRCCOPY, 
+    if (yChar >= 16) {
+      DrawBitmap(hdc, vgafont[cChar], prev_block_cursor_x*8,
+	              prev_block_cursor_y*16 + bx_headerbar_y, SRCCOPY,
 				  new_text[((prev_block_cursor_y*80 + prev_block_cursor_x)*2)+1]);
+    } else {
+      DrawChar(hdc, cChar, prev_block_cursor_x*8,
+               prev_block_cursor_y*yChar + bx_headerbar_y,
+               new_text[((prev_block_cursor_y*80 + prev_block_cursor_x)*2)+1], 1, 0);
+    }
   }
 
   for (i=0; i<nchars*2; i+=2) {
@@ -706,8 +744,11 @@ void bx_gui_c::text_update(Bit8u *old_text, Bit8u *new_text,
 
       x = (i/2) % 80;
       y = (i/2) / 80;
-
-      DrawBitmap(hdc, vgafont[cChar], x*8, y*16 + bx_headerbar_y, SRCCOPY, new_text[i+1]);
+      if(yChar>=16) {
+        DrawBitmap(hdc, vgafont[cChar], x*8, y*16 + bx_headerbar_y, SRCCOPY, new_text[i+1]);
+      } else {
+        DrawChar(hdc, cChar, x*8, y*yChar + bx_headerbar_y, new_text[i+1], 1, 0);
+      }
     }
   }
 
@@ -715,13 +756,23 @@ void bx_gui_c::text_update(Bit8u *old_text, Bit8u *new_text,
   prev_block_cursor_y = cursor_y;
 
   // now draw character at new block cursor location in reverse
-  if ((cursor_y*80 + cursor_x) < nchars ) {
+  if (((cursor_y*80 + cursor_x) < nchars ) && (cs_start <= cs_end)) {
     cChar = new_text[(cursor_y*80 + cursor_x)*2];
-	//reverse background and foreground colors
-	char cAttr = new_text[((cursor_y*80 + cursor_x)*2)+1];
-	cAttr = ((cAttr>>4)&0xf)+((cAttr&0xf)<<4);
-    DrawBitmap(hdc, vgafont[cChar], cursor_x*8, cursor_y*16 + bx_headerbar_y,
-	           SRCCOPY, cAttr);
+    if (yChar>=16)
+    {
+      memset(data, 0, sizeof(data));
+      for (unsigned i=0; i<16; i++) {
+        data[i*2] = reverse_bitorder(bx_vgafont[cChar].data[i]);
+        if ((i >= cs_start) && (i <= cs_end))
+          data[i*2] = 255 - data[i*2];
+      }
+      SetBitmapBits(cursorBmp, 32, data);
+      DrawBitmap(hdc, cursorBmp, cursor_x*8, cursor_y*16 + bx_headerbar_y,
+	         SRCCOPY, new_text[((cursor_y*80 + cursor_x)*2)+1]);
+    } else {
+      char cAttr = new_text[((cursor_y*80 + cursor_x)*2)+1];
+      DrawChar(hdc, cChar, cursor_x*8, cursor_y*yChar + bx_headerbar_y, cAttr, cs_start, cs_end);
+    }
   }
 
   ReleaseDC(stInfo.hwnd, hdc);
@@ -729,6 +780,43 @@ void bx_gui_c::text_update(Bit8u *old_text, Bit8u *new_text,
   LeaveCriticalSection(&stInfo.drawCS);
 }
 
+  int
+bx_gui_c::get_clipboard_text(Bit8u **bytes, Bit32s *nbytes)
+{
+  if (OpenClipboard(stInfo.hwnd)) {
+    HGLOBAL hg = GetClipboardData(CF_TEXT);
+    char *data = (char *)GlobalLock(hg);
+    *nbytes = strlen(data);
+    *bytes = (Bit8u *)malloc (*nbytes+1);
+    BX_INFO (("found %d bytes on the clipboard", *nbytes));
+    memcpy (*bytes, data, *nbytes+1);
+    BX_INFO (("first byte is 0x%02x", *bytes[0]));
+    GlobalUnlock(hg);
+    CloseClipboard();
+    return 1;
+  } else {
+    BX_ERROR (("paste: could not open clipboard"));
+    return 0;
+  }
+}
+
+  int
+bx_gui_c::set_clipboard_text(char *text_snapshot, Bit32u len)
+{
+  if (OpenClipboard(stInfo.hwnd)) {
+    HANDLE hMem = GlobalAlloc(GMEM_ZEROINIT, len);
+    EmptyClipboard();
+    lstrcpy((char *)hMem, text_snapshot);
+    SetClipboardData(CF_TEXT, hMem);
+    CloseClipboard();
+    GlobalFree(hMem);
+    return 1;
+  } else {
+    BX_ERROR (("copy: could not open clipboard"));
+    return 0;
+  }
+}
+
 
 // ::PALETTE_CHANGE()
 //
@@ -809,7 +897,22 @@ if ( x==dimension_x && y+bx_headerbar_y==dimension_y)
     stretched_y *= 2;
     stretch_factor *= 2;
   }
-  
+
+  FontId = 2;
+  yChar = 16;
+  if(y>600)
+  {
+    FontId = 0;
+    yChar = 12;
+    dimension_y = y * yChar / 16 + bx_headerbar_y;
+    stretched_y = dimension_y * stretch_factor;
+  } else if (y>480) {
+    FontId = 1;
+    yChar = 14;
+    dimension_y = y * yChar / 16 + bx_headerbar_y;
+    stretched_y = dimension_y * stretch_factor;
+  }
+
   SetWindowPos(stInfo.hwnd, HWND_TOP, 0, 0, stretched_x + x_edge * 2,
               stretched_y+ y_edge * 2 + y_caption,
                SWP_NOMOVE | SWP_NOZORDER);
@@ -844,7 +947,8 @@ unsigned bx_gui_c::create_bitmap(const unsigned char *bmap, unsigned xdim,
   for (unsigned i=0; i<ydim * xdim/8; i++)
     data[i] = reverse_bitorder(bmap[i]);
   SetBitmapBits(bx_bitmaps[bx_bitmap_entries].bmap, ydim * xdim/8, data);
-  free(data);
+  delete [] data;
+  data = NULL;
 
   bx_bitmaps[bx_bitmap_entries].xdim = xdim;
   bx_bitmaps[bx_bitmap_entries].ydim = ydim;
@@ -989,12 +1093,12 @@ void create_vga_font(void) {
 
 unsigned char reverse_bitorder(unsigned char b) {
   unsigned char ret=0;
-	
+
   for (unsigned i=0; i<8; i++) {
     ret |= (b & 0x01) << (7-i);
     b >>= 1;
   }
-	
+
   return(ret);
 }
 
@@ -1033,23 +1137,25 @@ void DrawBitmap (HDC hdc, HBITMAP hBitmap, int xStart, int yStart,
 //how to implement that so for now it's just implemented as color.
 //Note: it is also possible to program the VGA controller to have the
 //high bit for the foreground color enable blinking characters.
-	const COLORREF crPal[16] = { 
-									RGB(0 ,0 ,0 ),	//0 black 
-									RGB(0 ,0 ,0x80 ),	//1 dark blue 
-									RGB(0 ,0x80 ,0 ),	//2 dark green 
-									RGB(0 ,0x80 ,0x80 ),		//3 dark cyan 
-									RGB(0x80 ,0 ,0 ),	//4 dark red 
-									RGB(0x80 ,0 ,0x80 ),		//5 dark magenta 
-									RGB(0x80 ,0x80 ,0 ),		//6 brown
-									RGB(0xC0 ,0xC0 ,0xC0 ),	//7 light gray 
-									RGB(0x80 ,0x80 ,0x80 ),	//8 dark gray 
-									RGB(0 ,0 ,0xFF ),	//9 light blue 
-									RGB(0 ,0xFF ,0 ),	//10 green 
-									RGB(0 ,0xFF ,0xFF ),		//11 cyan 
-									RGB(0xFF ,0 ,0 ),	//12 light red 
-									RGB(0xFF ,0 ,0xFF ),		//13 magenta 
-									RGB(0xFF ,0xFF ,0 ),		//14 yellow 
-									RGB(0xFF ,0xFF ,0xFF )};	//15 white 
+
+	const COLORREF crPal[16] = {
+	RGB(0 ,0 ,0 ), //0 black 
+	RGB(0 ,0 ,0xA8 ), //1 dark blue 
+	RGB(0 ,0xA8 ,0 ), //2 dark green 
+	RGB(0 ,0xA8 ,0xA8 ), //3 dark cyan 
+	RGB(0xA8 ,0 ,0 ), //4 dark red 
+	RGB(0xA8 ,0 ,0xA8 ), //5 dark magenta 
+	RGB(0xA8 ,0x54 ,0 ), //6 brown 
+	RGB(0xA8 ,0xA8 ,0xA8 ), //7 light gray 
+	RGB(0x54 ,0x54 ,0x54 ), //8 dark gray 
+	RGB(0x54 ,0x54 ,0xFC ), //9 light blue 
+	RGB(0x54 ,0xFC ,0x54 ), //10 green 
+	RGB(0x54 ,0xFC ,0xFC ), //11 cyan 
+	RGB(0xFC ,0x54 ,0x54 ), //12 light red 
+	RGB(0xFC ,0x54 ,0xFC ), //13 magenta 
+	RGB(0xFC ,0xFC ,0x54 ), //14 yellow 
+	RGB(0xFC ,0xFC ,0xFC ) //15 white 
+	};
 
 	COLORREF crFore = SetTextColor(MemoryDC, crPal[(cColor>>4)&0xf]);
 	COLORREF crBack = SetBkColor(MemoryDC, crPal[cColor&0xf]);
@@ -1116,3 +1222,125 @@ void alarm (int time)
 }
 #endif
 #endif
+
+  void
+bx_gui_c::mouse_enabled_changed_specific (Boolean val)
+{
+}
+
+void DrawChar (HDC hdc, unsigned char c, int xStart, int yStart,
+               unsigned char cColor, int cs_start, int cs_end) {
+  HDC hdcMem;
+  POINT ptSize, ptOrg;
+  HGDIOBJ oldObj;
+  char str[2];
+  HFONT hFontOld;
+
+  hdcMem = CreateCompatibleDC (hdc);
+  SetMapMode (hdcMem, GetMapMode (hdc));
+  ptSize.x = 8;
+  ptSize.y = yChar;
+
+  DPtoLP (hdc, &ptSize, 1);
+
+  ptOrg.x = 0;
+  ptOrg.y = 0;
+
+  DPtoLP (hdcMem, &ptOrg, 1);
+
+  oldObj = SelectObject(MemoryDC, MemoryBitmap);
+  hFontOld=(HFONT)SelectObject(MemoryDC, hFont[FontId]);
+
+//Colors taken from Ralf Browns interrupt list.
+//(0=black, 1=blue, 2=red, 3=purple, 4=green, 5=cyan, 6=yellow, 7=white)
+//The highest background bit usually means blinking characters. No idea
+//how to implement that so for now it's just implemented as color.
+//Note: it is also possible to program the VGA controller to have the
+//high bit for the foreground color enable blinking characters.
+
+  const COLORREF crPal[16] = {
+  RGB(0 ,0 ,0 ), //0 black
+  RGB(0 ,0 ,0x80 ), //1 dark blue
+  RGB(0 ,0x80 ,0 ), //2 dark green
+  RGB(0 ,0x80 ,0x80 ), //3 dark cyan
+  RGB(0x80 ,0 ,0 ), //4 dark red
+  RGB(0x80 ,0 ,0x80 ), //5 dark magenta
+  RGB(0x80 ,0x80 ,0 ), //6 brown
+  RGB(0xc0 ,0xc0 ,0xc0 ), //7 light gray
+  RGB(0x80 ,0x80 ,0x80 ), //8 dark gray
+  RGB(0x00 ,0x00 ,0xff ), //9 light blue
+  RGB(0x00 ,0xff ,0x00 ), //10 green
+  RGB(0x00 ,0xff ,0xff ), //11 cyan
+  RGB(0xff ,0x00 ,0x00 ), //12 light red
+  RGB(0xff ,0x00 ,0xff ), //13 magenta
+  RGB(0xff ,0xff ,0x00 ), //14 yellow
+  RGB(0xff ,0xff ,0xff ) //15 white
+  };
+
+  COLORREF crFore = SetTextColor(MemoryDC, crPal[cColor&0xf]);
+  COLORREF crBack = SetBkColor(MemoryDC, crPal[(cColor>>4)&0xf]);
+  str[0]=c;
+  str[1]=0;
+
+  int y = FontId == 2 ? 16 : 8;
+
+  TextOut(MemoryDC, xStart, yStart, str, 1);
+  if (cs_start <= cs_end && cs_start < y)
+  {
+    RECT rc;
+    SetBkColor(MemoryDC, crPal[cColor&0xf]);
+    SetTextColor(MemoryDC, crPal[(cColor>>4)&0xf]);
+    rc.left = xStart+0;
+    rc.right = xStart+8;
+    if (cs_end >= y)
+      cs_end = y-1;
+    rc.top = yStart+cs_start*yChar/y;
+    rc.bottom = yStart+(cs_end+1)*yChar/y;
+    ExtTextOut(MemoryDC, xStart, yStart, ETO_CLIPPED|ETO_OPAQUE, &rc, str, 1, NULL);
+  }
+
+  SetBkColor(MemoryDC, crBack);
+  SetTextColor(MemoryDC, crFore);
+
+  SelectObject(MemoryDC, hFontOld);
+  SelectObject(MemoryDC, oldObj);
+
+  updateUpdated(xStart, yStart, ptSize.x + xStart - 1, ptSize.y + yStart - 1);
+
+  DeleteDC (hdcMem);
+}
+
+void InitFont(void)
+{
+  LOGFONT lf;
+  int i;
+
+  lf.lfWidth = 8;
+  lf.lfEscapement = 0;
+  lf.lfOrientation = 0;
+  lf.lfWeight = FW_MEDIUM;
+  lf.lfItalic = FALSE;
+  lf.lfUnderline=FALSE;
+  lf.lfStrikeOut=FALSE;
+  lf.lfCharSet=OEM_CHARSET;
+  lf.lfOutPrecision=OUT_DEFAULT_PRECIS;
+  lf.lfClipPrecision=CLIP_DEFAULT_PRECIS;
+  lf.lfQuality=DEFAULT_QUALITY;
+  lf.lfPitchAndFamily=FIXED_PITCH | FF_DONTCARE;
+  wsprintf(lf.lfFaceName, "Lucida Console");
+
+  for (i=0; i < 3; i++)
+  {
+    lf.lfHeight = 12 + i * 2;
+    hFont[i]=CreateFontIndirect(&lf);
+  }
+}
+
+void DestroyFont(void)
+{
+  int i;
+  for(i = 0; i < 3; i++)
+  {
+    DeleteObject(hFont[i]);
+  }
+}
